package org.zeroturnaround.javarebel.integration.spring;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.springframework.beans.factory.xml.ParserContext;
import org.springframework.context.annotation.ComponentScanBeanDefinitionParser;
import org.w3c.dom.Element;
import org.zeroturnaround.javarebel.LoggerFactory;

/**
 * Requires Spring 2.1 M3 or later.
 * <p>
 * Should only be used if class
 * org.springframework.context.annotation.ComponentScanBeanDefinitionParser
 * exists.
 */
public class AnnotationBeanDefinitionManager {
  private static List componentScanners = 
    Collections.synchronizedList(new ArrayList());
  private static Method reparseMethod;
  
  private static long lastScan = 0;
  
  static {
    try {
      reparseMethod = ComponentScanBeanDefinitionParser.class.getDeclaredMethod(
          "reparse", new Class[] {
              Element.class,
              ParserContext.class});
    } catch (SecurityException e) {
      LoggerFactory.getInstance().errorEcho(e);
    } catch (NoSuchMethodException e) {
      LoggerFactory.getInstance().errorEcho(e);
    }
  }
  
  private static class ComponentScanner {
    private ComponentScanBeanDefinitionParser componentScanner; 
    private Element element;
    private ParserContext parserContext;
    
    public ComponentScanner(ComponentScanBeanDefinitionParser componentScanner,
        Element element, ParserContext parserContext) {
      this.componentScanner = componentScanner;
      this.element = element;
      this.parserContext = parserContext;
    }
    
    public void scan() throws Throwable {
      try {
        reparseMethod.invoke(componentScanner, new Object[] {element, parserContext});
      } catch (IllegalAccessException e) {
        LoggerFactory.getInstance().errorEcho(e);
        throw e;
      } catch (InvocationTargetException e) {
        LoggerFactory.getInstance().errorEcho(e.getTargetException());
        throw e.getTargetException();
      }
    }
  }
  
  public static void registerComponentScanner(
      ComponentScanBeanDefinitionParser componentScanner, 
      Element element, 
      ParserContext parserContext) {
    componentScanners.add(new ComponentScanner(componentScanner, element, parserContext));
  }
  
  public static void rescan() {
    if (lastScan + 1000 > System.currentTimeMillis())
      return;       
    
    try {
      for (Iterator i = new ArrayList(componentScanners).iterator(); i.hasNext();) {
        ComponentScanner cs = (ComponentScanner) i.next();
        cs.scan();
      }
    } 
    catch (Throwable e) {
      if (e instanceof Error)
        throw (Error) e;
      if (e instanceof RuntimeException)
        throw (RuntimeException) e;
      throw new RuntimeException(e);
    }
    
    lastScan = System.currentTimeMillis();
  }
}
